#ifndef MPP_DRAW_H_
#define MPP_DRAW_H_

#include <utility>
#include <deque>
#include <algorithm>
#include <vector>
#include <string>
#include <sstream>
#include <iostream>

#include <FL/Fl_Widget.H>
#include <FL/Enumerations.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Box.H>

#include <mpp/mpp.h>
#include <mpp/scale.h>

namespace mpp
{
extern Fl_Window *draw_window;
inline void set_color(Fl_Color color) {
    fl_color(color);
}
// Primitives
inline void point(Float x, Float y);
inline void point(Float x, Float y, Fl_Color color);
inline void pointb(Float x, Float y);
inline void pointb(Float x, Float y, Fl_Color color);
inline void line(Float x, Float y, Float x1, Float y1);
inline void line(Float x, Float y, Float x1, Float y1, Fl_Color color);
inline void circle(Float x, Float y, Float r);
inline void circle(Float x, Float y, Float r, Fl_Color color);
inline void circlef(Float x, Float y, Float r);

//Attributes
extern int point_size;
extern int point_size_b;

class Shape
{
public:
    virtual void draw()const =0;
    virtual ~Shape() {};
protected:
};

class User_draw: public Shape
{
public:
    inline User_draw(void (*ff)());
    inline User_draw(void (*ff)(void*), void *);
    inline void draw()const;
private:
    void (*ff)();
    void (*ff2)(void*);
    void * obj;
};

class Circle: public Shape
{
public:
    double x, y, R;
    Fl_Color color;
    const Scale * scale;
    Circle(Float x, Float y, Float R, Fl_Color, const Scale *);
    //inline Circle(Float x, Float y, Float R, Fl_Color, const Scale &);
    void draw()const;
};

class Hist: public Shape
{
private:
    //unsigned int nc;
    double xmin;
    double xmax;
    unsigned int ymax;
    int legs;
    std::vector<unsigned int> chan;
    Fl_Color color;
    const Scale *scale;
public:
    Hist(unsigned int nc_, double xmin_, double xmax_, Fl_Color color1, const Scale * scale_, int legs_=0);
    //        Hist(unsigned int nc_, double xmin_, double xmax_, Fl_Color color1, const Scale & scale_, int legs_=0);
    void add(const double *x, unsigned int sz);
    void set(const double *x, unsigned int sz);
    void draw()const;
};

class Grid: public Shape
{
private:
public:
    inline Grid(double xmin, double dx, double ymin, double dy, const Scale * scale);
    double xmin, dx, ymin, dy;
    inline void draw()const;
    const Scale *scale;
};

inline std::ostream & operator<<(std::ostream & os, const Grid& grid) {
    return os << grid.xmin << ' ' << grid.dx << ' ' << grid.ymin << ' ' << grid.dy;
}

inline std::istream & operator>>(std::istream & is, Grid& grid) {
    return is >> grid.xmin >> grid.dx >> grid.ymin >> grid.dy;
}

template <class T> class Plot: public Shape
{
public:
    Plot(T x1, T x2, T y1, Fl_Color color, const Scale *scale):scale(scale),color(color),x1(x1),x2(x2),y1(y1) {};
    //		Plot(T x1, T x2, T y1, Fl_Color color, const Scale &scale):scale(&scale),color(color),x1(x1),x2(x2),y1(y1){};
    void draw()const
    {
        set_scale(scale);
        set_color(color);
        T x=x1, y=y1;
        while(x != x2) point(*x++, *y++);
    };
    const Scale *scale;
    Fl_Color color;
private:
    const T x1, x2, y1;
};

class Roll: public Shape
{
private:
    std::deque<std::pair<Float, Float> > roll_q;
    Fl_Color color;
    const Scale *scale;
public:
    Roll(Fl_Color color1, const Scale *s);
    //        Roll(Fl_Color color1, const Scale &s);
    void add(double x, double y);
    void draw()const;
    void reset();
};

class Orbit: public Shape
{
private:
    std::vector<std::pair<Float, Float> > pvec;
    unsigned int sz_all;
    unsigned int sz_head;// tail size
    unsigned int p1, p2; // position of last point
    const Scale *scale;
public:
    Fl_Color color_head; // head color
    Fl_Color color_tail; // tail color
    Orbit(int sz_all, int sz_head, Fl_Color color_head, Fl_Color color_tail, const Scale*);
    //        Orbit(int sz_all, int sz_head, Fl_Color color_head, Fl_Color color_tail, const Scale&);
    void draw()const;
    void add(double x, double y);
    void reset();
};

// begin iter, end iter, values iter
template <class T> void plot(T x1, T x2, T y1);
// f(double)
template <class T> void plot(const T& f, double xl, double xr, int n);
void plot(double(*)(double), double xl, double xr, int n);
//template<int w, int h> void plot(double[w][h])

// Implementation
void point(Float x, Float y) {
    fl_rectf(transx(x)-point_size, transy(y)-point_size, 1+2*point_size, 1+2*point_size);
}
void point(Float x, Float y, Fl_Color color) {
    fl_color(color);
    point(x,y);
}
void pointb(Float x, Float y) {
    fl_rectf(transx(x)-point_size_b, transy(y)-point_size_b, 1+2*point_size_b, 1+2*point_size_b);
}
void pointb(Float x, Float y, Fl_Color color) {
    fl_color(color);
    pointb(x,y);
}
void line(Float x, Float y, Float x1, Float y1) {
    fl_line(transx(x), transy(y), transx(x1), transy(y1));
}
void line(Float x, Float y, Float x1, Float y1, Fl_Color color) {
    fl_color(color);
    fl_line(transx(x), transy(y), transx(x1), transy(y1));
}
void circle(Float x, Float y, Float r) {
    fl_circle(transx(x), transy(y), scalex(r));
}
void circle(Float x, Float y, Float r, Fl_Color color) {
    fl_color(color);
    circle(x,y,r);
}
void circlef(Float x, Float y, Float r) {
    fl_pie(transx(x-r), transy(y+r), scalex(2*r), scalex(2*r), 0, 360);
}

//    Circle::Circle(Float x, Float y, Float R, Fl_Color color, const Scale & scale):
//        x(x), y(y), R(R), color(color), scale(&scale){}

User_draw::User_draw(void (*ff)() ):ff(ff),ff2(0),obj(0) {};
User_draw::User_draw(void (*ff2)(void*), void *obj):ff(0),ff2(ff2),obj(obj) {};
void User_draw::draw()const
{
    if (ff) ff();
    if (ff2) ff2(obj);
}

template <class T> void plot(T x1, T x2, T y1) {
    for(; x1!=x2; ++x1, ++y1) point(*x1, *y1);
}

template <class T> void plot(const T & f, double xl, double xr, int n) {
    double xx=(xr-xl)/(n-1);
    for (int i=0; i<n; ++i) {
        point(xl+xx*i, f(xl+xx*i));
    }
}

class Bitmap: public Shape
{
private:
    mutable int W, H;
    int w, h;
    double c1, c2;
    mutable uchar *map;
    const double *rmap;
    const Scale *sc;
public:
    Bitmap(const double *, int w, int h, double C1, double C2, const Scale *);
    //		Bitmap(const double *, int w, int h, double C1, double C2, const Scale &scale);
    ~Bitmap();
    void draw()const;
};


template<typename T>class BitmapRB: public Shape
{
private:
    mutable int W, H;
    int w, h;
    T c1, c2;
    mutable uchar *map;
    const T *rmap;
    const T *bmap;
    const Scale *sc;
public:
    BitmapRB(const T *rmap1, const T *bmap1, int w1, int h1, T c1, T c2, const Scale *sc1):c1(c1), c2(c2) {
        sc=sc1;
        rmap=rmap1;
        bmap=bmap1;
        w=w1;
        h=h1;
        map = 0;
        W=H=0;
    }
    ~BitmapRB() {
        delete[] map;
    }
    void draw()const {
        set_scale(sc);
        if ( ( (int)(scale_.X*w) !=W ) || ( (int)(scale_.Y*h) != H) ) {
            H=abs((int)(scale_.Y*h));
            W=abs((int)(scale_.X*w));
            delete[] map;
            map = new uchar[H*W*4];
        }
        double tmp=255.0/(c2-c1);
        for (int i=0; i<H; i++)
            for(int j=0; j<W; j++) {
                int red_ =255*(rmap[w*(int)(i*(1.0*h)/H)+(int)(j*(1.0*w)/W)]-c1)/(c2-c1);
                int blue_=255*(bmap[w*(int)(i*(1.0*h)/H)+(int)(j*(1.0*w)/W)]-c1)/(c2-c1);
                //int green_=
                map[(i*W+j)*4]=(uchar)(255-blue_);
                map[(i*W+j)*4+1]=(uchar)(255- (blue_>red_? blue_ : red_) );
                map[(i*W+j)*4+2]=(uchar)(255-red_);
                map[(i*W+j)*4+3]=(uchar)(0);
            }

        fl_draw_image(map, transx(0), transy(0), W, H, 4, 0);
    }
};



// Draws one array relative to another
class Plot_a: public Shape
{
private:
    const double *x, *y;
    int sz;
public:
    Plot_a(const double *x, const double *y, int sz, Fl_Color color, const Scale & scale):x(x),y(y),sz(sz),color(color),scale(&scale) {};
    Fl_Color color;
    const Scale *scale;
    void draw()const;
};

// Draws one vector relative to another
class Plot_v: public Shape
{
private:
    const std::vector<double> &x;
    const std::vector<double> &y;
public:
    Plot_v(const std::vector<double> &x, const std::vector<double> &y, Fl_Color color, const Scale scale):x(x), y(y),color(color),scale(&scale) {};
    Fl_Color color;
    const Scale *scale;
    void draw()const;
};

// Draws list of points. See std::pair
class Plot_l: public Shape
{
private:
    const std::list<std::pair<mpp::Float, mpp::Float> > &pointl;
public:
    Fl_Color color;
    const Scale *scale;
    Plot_l(const std::list<std::pair<mpp::Float, mpp::Float> > &pointl, Fl_Color color, const Scale &scale):pointl(pointl),color(color),scale(&scale) {};
    void draw()const;
};
};

// Implementation

mpp::Grid::Grid(double xmin, double dx, double ymin, double dy, const Scale * scale):
    xmin(xmin), dx(dx), ymin(ymin), dy(dy), scale(scale) {}

void mpp::Grid::draw()const
{
    set_scale(scale);
    fl_color(FL_DARK_GREEN);
    double x=xmin, y=ymin;
    fl_font(0,18);
    std::stringstream ss;
    std::string str;
    ss<<xmin<<std::endl;
    ss>>str;
    fl_draw(str.c_str(), transx(x)+5, transy(y)-10, 0, 0, FL_ALIGN_LEFT);
    while(!(x>scale->x2)) {
        mpp::line(x, scale->y1, x, scale->y2);
        x+=dx;
    }
    ss<<(x-=dx)<<std::endl;
    ss>>str;
    fl_draw(str.c_str(), transx(x), transy(y)-10, 0, 0, FL_ALIGN_RIGHT);
    ss<<y<<std::endl;
    ss>>str;
    fl_draw(str.c_str(), transx(xmin)+5, transy(y)-30, 0, 0, FL_ALIGN_LEFT);
    while(!(y>scale->y2)) {
        mpp::line(scale->x1, y, scale->x2, y);
        y+=dy;
    }
    ss<<(y-=dy)<<std::endl;
    ss>>str;
    fl_draw(str.c_str(), transx(xmin)+5, transy(y)+10, 0, 0, FL_ALIGN_LEFT);
};
#endif
